package blueeyes.io.iteratee

import scalaz.Scalaz._
import scalaz.{Validation, Success, Failure, Identity, Pure, Semigroup}

trait Experiment {
  trait Mealy[A, B, M] extends PartialFunction[A, (B, M)]

  /** A step in the life of a finite mealy machine.
   */
  sealed trait Step[A] {
    /** The value emitted at this step.
     */
    def value: A

    def map[AA](f: A => AA): Step[AA]

    def fold[Z](continue: A => Z, done: A => Z): Z
  }
  /** A step indicating the finite mealy machine can process more input.
   */
  sealed case class Continue[A](value: A) extends Step[A] {
    def map[AA](f: A => AA): Step[AA] = copy(value = f(value))

    def fold[Z](continue: A => Z, done: A => Z): Z = continue(value)
  }
  /** A step indicating the finite mealy machine cannot process any more input.
   */
  sealed case class Done[A](value: A) extends Step[A] {
    def map[AA](f: A => AA): Step[AA] = copy(value = f(value))

    def fold[Z](continue: A => Z, done: A => Z): Z = done(value)
  }

  trait FiniteMealy[A, B, M] extends Mealy[A, Step[B], M]

  implicit def InputSemigroup[A](implicit sga: Semigroup[A]): Semigroup[Input[A]] = new Semigroup[Input[A]] {
    def append(i1: Input[A], i2: => Input[A]): Input[A] = (i1, i2) match {
      case (Input.Chunk(elem1), Input.Chunk(elem2)) => Input.Chunk(sga.append(elem1, elem2))

      case (_, elem2: Input.Chunk[_]) => elem2

      case (elem1: Input.Chunk[_], _) => elem1

      case (_, input) => input
    }
  }

  implicit def inputToLhsIterateeRemainderAdder[A: Semigroup](input1: Input[A]) = new {
    /** Prepends this remainder to the iteratee on the RHS.
     */
    def +\ [B](iter: Iteratee[A, B]): Iteratee[A, B] = Iteratee.iteratee { (input2: Input[A]) =>
      val input = implicitly[Semigroup[Input[A]]].append(input1, input2)

      iter(input)
    }
  }

  sealed trait IterateeOutput[+A, +B] {
    def isEmpty: Boolean

    def getOrElse[BB >: B](v: => BB): BB

    def map[BB](f: B => BB): IterateeOutput[A, BB]

    def filter(f: B => Boolean): IterateeOutput[A, B]

    def orElse[AA >: A, BB >: B](v: => BB, r: => Input[AA]): IterateeOutput[AA, BB]

    def orElse[AA >: A, BB >: B](alt: => IterateeOutput[AA, BB]): IterateeOutput[AA, BB]

    def /+ [AA >: A](r2: Input[AA])(implicit sg: Semigroup[Input[AA]]): IterateeOutput[AA, B]

    def /-> [AA >: A](r: Input[AA]): IterateeOutput[AA, B]
  }
  case object Consume extends IterateeOutput[Nothing, Nothing] {
    def isEmpty: Boolean = true

    def getOrElse[BB](v: => BB): BB = v

    def map[BB](f: Nothing => BB): IterateeOutput[Nothing, Nothing] = Consume

    def filter(f: Nothing => Boolean): IterateeOutput[Nothing, Nothing] = Consume

    def orElse[AA, BB](v: => BB, r: => Input[AA]): IterateeOutput[AA, BB] = Produce(v, r)

    def orElse[AA, BB](alt: => IterateeOutput[AA, BB]): IterateeOutput[AA, BB] = alt

    def /+ [AA](r2: Input[AA])(implicit sg: Semigroup[Input[AA]]): IterateeOutput[Nothing, Nothing] = Consume

    def /-> [AA](r: Input[AA]): IterateeOutput[Nothing, Nothing] = Consume
  }
  case class Produce[+A, +B](value: B, remainder: Input[A]) extends IterateeOutput[A, B] {
    def isEmpty: Boolean = false

    def getOrElse[BB >: B](v: => BB): BB = value

    def map[BB](f: B => BB): IterateeOutput[A, BB] = copy(value = f(value))

    def filter(f: B => Boolean): IterateeOutput[A, B] = if (f(value)) this else Consume

    def orElse[AA >: A, BB >: B](v: => BB, r: => Input[AA]): IterateeOutput[AA, BB] = this

    def orElse[AA >: A, BB >: B](alt: => IterateeOutput[AA, BB]): IterateeOutput[AA, BB] = this

    def /+ [AA >: A](r2: Input[AA])(implicit sg: Semigroup[Input[AA]]): IterateeOutput[AA, B] = copy(remainder = sg.append(remainder, r2))

    def /-> [AA >: A](r: Input[AA]): IterateeOutput[AA, B] = copy(remainder = r)
  }

  type IterateeError = Exception

  type IterateeState[A, B] = (Step[Validation[IterateeError, IterateeOutput[A, B]]], Iteratee[A, B])
  type IterateeStateFactory[A, B] = Iteratee[A, B] => IterateeState[A, B]

  sealed trait Iteratee[A, B] extends FiniteMealy[Input[A], Validation[IterateeError, IterateeOutput[A, B]], Iteratee[A, B]] { self =>
    import Iteratee._

    def isDefinedAt(input: Input[A]): Boolean = true

    /** Appends additional input to the final remainder output by this iteratee.
     *
     * In the below example, when `iteratee` is done, any remainder it outputs
     * will be augmented with the specified additional input.
     *
     * {{{
     * iteratee /+ extraInput
     * }}}
     *
     */
    def /+ (leftOver: Input[A])(implicit sga: Semigroup[A]):Iteratee[A, B] = iteratee { (input: Input[A]) =>
      val (state, next) = self(input)

      state match {
        case Done(Success(Produce(value, remainder))) =>
          (state.map(_.map(_ /+ leftOver)), next)

        case _ => (state, next /+ leftOver)
      }
    }

    /** Removes all values not accepted by the filter.
     */
    final def filter(f: B => Boolean): Iteratee[A, B] = new Iteratee[A, B] {
      override def isDefinedAt(input: Input[A]): Boolean = self.isDefinedAt(input)

      def apply(input: Input[A]): IterateeState[A, B] = {
        val (state, next: Iteratee[A, B]) = self.apply(input)

        state match {
          case Done(Success(Produce(value, remainder))) =>
            (state.map(_.map(_.filter(f))), next.filter(f))

          case Continue(Success(Produce(value, remainder))) =>
            (state.map(_.map(_.filter(f) /-> Input.Empty)), next.filter(f))

          case _ =>
            (state.map(_.map(_.filter(f))), next.filter(f))
        }
      }
    }

    /** Maps the iteratee into another one that produces values of a different
     * type, but in 1-to-1 correspondance with this type.
     */
    final def map[BB](f: B => BB): Iteratee[A, BB] = new Iteratee[A, BB] {
      override def isDefinedAt(input: Input[A]): Boolean = self.isDefinedAt(input)

      def apply(input: Input[A]): IterateeState[A, BB] = {
        val (step, next: Iteratee[A, B]) = self.apply(input)

        (step.map(_.map(_.map(f))), next.map(f))
      }
    }

    /** Returns the iteratee resulting from producing a single value from this
     * iteratee and feeding it to the specified function.
     *
     * If this iteratee produces many values, and you want to collect all of them
     * before you flat map to another iteratee, then you should first call the
     * collectAll method.
     */
    final def flatMap[BB](f: B => Iteratee[A, BB]): Iteratee[A, BB] = {
      def flatMap0(iter: Iteratee[A, B]): Iteratee[A, BB] = iteratee { (input: Input[A]) =>
        val (state, next) = iter(input)

        state match {
          case Continue(Success(Produce(value, remainder))) => f(value)(remainder)
          case Continue(Success(Consume))                   => continue(flatMap0(next))

          case Done(Success(Produce(value, remainder)))     => f(value)(remainder)
          case Done(Success(Consume))                       => throwErrorFatal(new Exception("Cannot flatMap on iterator that produces no values!"))

          case Continue(Failure(err))                       => throwError(err)(flatMap0(next))
          case Done(Failure(err))                           => throwErrorFatal(err)
        }
      }

      flatMap0(self)
    }

    final def >>= [BB](f: B => Iteratee[A, BB]): Iteratee[A, BB] = flatMap(f)

    final def then[BB](that: Iteratee[A, BB]): Iteratee[A, BB] = flatMap(_ => that)

    final def >> [BB](that: Iteratee[A, BB]): Iteratee[A, BB] = self.then(that)

    /**
     * {{{
     * val iter1: Iteratee[JValue, JValue]
     * val iter2: Iteratee[Array[Byte], JValue] = iter1 <<< byteArrayToJValue
     * }}}
     */
    final def <<< [AA](trans: Iteratee[AA, A]): Iteratee[AA, B] = trans >>> self

    /**
     * {{{
     * val iter1: Iteratee[JValue, JValue]
     * val iter2: Iteratee[JValue, Array[Byte]] = iter1 >>> jvalueToByteArray
     * val iter3: Iteratee[Array[Byte], Array[Byte]] = (iter1 >>> jvalueToByteArray) <<< byteArrayToJValue
     * }}}
     */
    final def >>> [C](trans: Iteratee[B, C]): Iteratee[A, C] = {
      def compose0(iter: Iteratee[A, B], trans: Iteratee[B, C]): Iteratee[A, C] = iteratee { (input: Input[A]) =>
        val (state, next) = iter(input)

        state match {
          case Continue(Success(Consume)) => continue(compose0(next, trans))

          case Continue(Success(Produce(value, remainder1))) => {
            val (transState, transNext) = trans(Input.Chunk(value))

            transState match {
              case Continue(Success(Produce(value, _))) => continue(value, remainder1)(compose0(next, transNext))
              case Continue(Success(Consume))           => continue(compose0(next, transNext))
              case Continue(Failure(err))               => throwError(err)(compose0(next, transNext))

              case Done(Success(Produce(value, _)))     => done(value, remainder1)
              case Done(Success(Consume))               => done
              case Done(Failure(err))                   => throwErrorFatal(err)
            }
          }

          case Continue(Failure(err)) => throwError(err)(compose0(next, trans))

          case Done(Success(Consume)) => done

          case Done(Success(Produce(value, remainder1))) => {
            val (transState, transNext) = trans(Input.Chunk(value))

            transState match {
              case Continue(Success(Produce(value, _))) => done(value, remainder1)
              case Continue(Success(Consume))           => done
              case Continue(Failure(err))               => throwErrorFatal(err)

              case Done(Success(Produce(value, _)))     => done(value, remainder1)
              case Done(Success(Consume))               => done
              case Done(Failure(err))                   => throwErrorFatal(err)
            }
          }

          case Done(Failure(err)) => throwErrorFatal(err)
        }
      }

      compose0(self, trans)
    }

    /** Selects the last value produced by the iteratee.
     */
    def last: Iteratee[A, B] = {
      def last0(iter: Iteratee[A, B], last: IterateeOutput[A, B]): Iteratee[A, B] = iteratee { (input: Input[A]) =>
        val nextStep = iter(input)

        val (state, next) = nextStep

        state match {
          case Continue(Success(value)) => continue(last0(next, value.orElse(last)))
          case Continue(Failure(err))   => throwError(err)(last0(next, last))

          case Done(Success(value))     => done(value.orElse(last))
          case Done(Failure(err))       => nextStep
        }
      }

      last0(self, Consume)
    }

    /** Scans over all values produced by the iteratee.
     */
    def scan[Z](initial: Z)(f: (Z, B) => Z): Iteratee[A, Z] = {
      def accumulate(acc: Z, value: IterateeOutput[A, B]): IterateeOutput[A, Z] = value.map(value => f(acc, value))

      def scan0(iter: Iteratee[A, B], acc: Z): Iteratee[A, Z] = iteratee { (input: Input[A]) =>
        val (state, next) = iter(input)

        state match {
          case Continue(Success(value)) =>
            val nextAcc = accumulate(acc, value)

            continue(nextAcc)(scan0(next, nextAcc.getOrElse(acc)))

          case Continue(Failure(err)) =>
            throwError(err)(scan0(next, acc))

          case Done(Success(value)) =>
            val nextAcc = accumulate(acc, value)

            done(nextAcc)

          case Done(Failure(err)) =>
            throwErrorFatal(err)
        }
      }

      scan0(self, initial)
    }

    /** Folds over all values produced by the iteratee.
     */
    def fold[Z](initial: Z)(f: (Z, B) => Z): Iteratee[A, Z] = scan(initial)(f).last

    /** Returns a derivative iteratee that guarantees the production of a value.
     * If this iteratee does not produce any value before it finishes, then the
     * specified default value is produced.
     *
     * Fatal errors in this iteratee are mapped to the default value. However,
     * recoverable errors in this iteratee are mapped to recoverable errors in
     * the returned iteratee.
     */
    def orElse(b: B): Iteratee[A, B] = {
      def orElse0(iter: Iteratee[A, B]): Iteratee[A, B] = iteratee { (input: Input[A]) =>
        val nextStep = iter(input)

        val (state, next) = nextStep

        state match {
          case Continue(Success(value)) =>
            if (!value.isEmpty) nextStep
            else continue(value)(orElse0(next))

          case Continue(Failure(err)) =>
            throwError(err)(orElse0(next))

          case Done(Success(value)) =>
            if (!value.isEmpty) nextStep
            else done(value.orElse(b, Input.Empty))

          case Done(Failure(err)) =>
            done(b, Input.Empty)
        }
      }

      orElse0(self)
    }

    /** Collects all values produced by the iteratee into a (possibly empty)
     * list.
     */
    def collectAll: Iteratee[A, List[B]] = (fold[List[B]](Nil) { (list, e) => e :: list}).map(_.reverse)
  }

  object Iteratee {
    def FatalErrorIteratee[A, B](err: IterateeError): Iteratee[A, B] = iteratee { (input: Input[A]) => (Done(failure(err)), FatalErrorIteratee(err)) }

    /** Creates an iteratee from the specified function.
     */
    def iteratee[A, B](f: Input[A] => IterateeState[A, B]): Iteratee[A, B] = new Iteratee[A, B] {
      def apply(input: Input[A]) = f(input)
    }

    def continue[A, B](output: IterateeOutput[A, B]): IterateeStateFactory[A, B] = (next: Iteratee[A, B]) => (Continue(success(output)), next)

    def continue[A, B](value: B, remainder: Input[A]): IterateeStateFactory[A, B] = continue(Produce(value, remainder))

    def continue[A, B](next: Iteratee[A, B]): IterateeState[A, B] = continue(Consume)(next)

    def done[A, B](output: IterateeOutput[A, B]): IterateeState[A, B] = (Done(success(output)), FatalErrorIteratee(AlreadyDone))

    def done[A, B](value: B, remainder: Input[A]): IterateeState[A, B] = done(Produce(value, remainder))

    def done[A, B]: IterateeState[A, B] = done(Consume)

    def throwError[A, B](err: IterateeError): IterateeStateFactory[A, B] = (next: Iteratee[A, B]) => (Continue(failure(err)), next)

    def throwErrorFatal[A, B](err: IterateeError): IterateeState[A, B] = (Done(failure(err)), FatalErrorIteratee(err))
  }

  sealed trait Enumerator[A] {
    def fold[B, Z](iter: Iteratee[A, B], state: Z, f: (Z, B) => Z): (Z, IterateeState[A, B])
  }
}
